
from .surface import Surface
from .body import Body
from .module import Module
from .clone import Clone
from .include import Include
from .end import End

__all__ = ["Surface", "Body", "Module", "Clone", "Include", "End"]


from . import base

class Module(base.AdvancedElement):
    """
    Creation of modules according to the next formatted text lines

    0000000000000000000000000000000000000000000000000000000000000000
    MODULE  (  A4) comment
    MATERIAL(  A4)
    SURFACE (  A4), SIDE POINTER=(I2)
    BODY    (  A4)
    MODULE  (  A4)
    1111111111111111111111111111111111111111111111111111111111111111
      OMEGA=(         E22.15       ,  I4) DEG (DEFAULT=0.0)
      THETA=(         E22.15       ,  I4) DEG (DEFAULT=0.0)
        PHI=(         E22.15       ,  I4) DEG (DEFAULT=0.0)
    X-SHIFT=(         E22.15       ,  I4) (DEFAULT=0.0)
    Y-SHIFT=(         E22.15       ,  I4) (DEFAULT=0.0)
    Z-SHIFT=(         E22.15       ,  I4) (DEFAULT=0.0)
    """

    def __init__(self, label, material, comment="", **kwargs):
        super().__init__(label, material, comment, **kwargs)
        
    def __str__(self):
        s="0000000000000000000000000000000000000000000000000000000000000000\n"
        s+="MODULE  ({0}) {1}\n".format(self.label, self.comment)
        # material
        s+="MATERIAL({0})".format(("    "+str(self.void_inner_volume_factor*self.material if self.material<0 else self.material))[-4:])
        # surfaces, bodies and modules 
        s+=self.representation_elements()
        # rotation and translation
        if self.rotation!=[0,0,0] or self.translation!=[0,0,0]:
            s+="\n1111111111111111111111111111111111111111111111111111111111111111"
        # rotation
        s+=self.representation_rotation()
        # translation
        s+=self.representation_translation()
        return s



from . import base

class Include(base.Base):
    """
    Insertion of others geometry definition files according to the next formatted text lines
    
    0000000000000000000000000000000000000000000000000000000000000000
    INCLUDE  comment
       FILE=(filename.ext)
       
    filename (string): absolute or relative path of the geometry definition file to be included
    starred (boolean): if True its allow the included file to be considered as if it were part of the main file
    comment (string): short text description
    """

    def __init__(self, filename, starred=False, comment="", **kwargs):
        super().__init__(comment)
        self.filename=filename
        self.starred=starred
        
    def __str__(self):
        s="0000000000000000000000000000000000000000000000000000000000000000\n"
        s+="INCLUDE{0} {1}\n".format("*" if self.starred else " ", self.comment)
        s+="   FILE={0}".format(self.filename)
        return s



from abc import ABC, abstractmethod


class Base(ABC):
    def __init__(self, comment=""):
        self.comment=comment
        self.void_inner_volume_factor=1
        
    @abstractmethod
    def __str__(self):
        pass
    
    def set_void_inner_volume_factor(self, factor):
        self.void_inner_volume_factor=factor
      
      
class Element(Base):
    def __init__(self, label, comment="", **kwargs):
        super().__init__(comment)
        self.label=("    "+label)[-4:]
        
        # length unit
        self.unit_multiplication_factor=1
        self.unit=kwargs.get("unit", "cm").lower()
        self.compute_unit_multiplication_factor()
        
        self.scale=list(kwargs.get("scale", [1,1,1]))
        self.rotation=list(kwargs.get("rotation", [0,0,0]))
        self.translation=list(kwargs.get("translation", [0,0,0]))
        
        # scale (single parameters)
        self.scale[0]=kwargs.get("xscale", self.scale[0])
        self.scale[1]=kwargs.get("yscale", self.scale[1])
        self.scale[2]=kwargs.get("zscale", self.scale[2])
        
        # rotation (single parameters)
        self.rotation[0]=kwargs.get("omega", self.rotation[0])
        self.rotation[1]=kwargs.get("theta", self.rotation[1])
        self.rotation[2]=kwargs.get("phi", self.rotation[2])
        
        # translation (single parameters)
        self.translation[0]=kwargs.get("xshift", self.translation[0])
        self.translation[1]=kwargs.get("yshift", self.translation[1])
        self.translation[2]=kwargs.get("zshift", self.translation[2])
        
        # angle unit
        self.angle=kwargs.get("angle", "DEG").upper()[:3]
        if self.angle!="RAD":
            self.angle="DEG"

    @abstractmethod
    def __str__(self):
        pass

    def compute_unit_multiplication_factor(self):
        if self.unit.lower() in ["mm", "millimeter", "millimeters"]:
            self.unit_multiplication_factor=0.1
        elif self.unit.lower() in ["cm", "centimeter", "centimeters"]:
            self.unit_multiplication_factor=1
        elif self.unit.lower() in ["dm", "decimeter", "decimeters"]:
            self.unit_multiplication_factor=10
        elif self.unit.lower() in ["m", "meter", "meters"]:
            self.unit_multiplication_factor=100
        elif self.unit.lower() in ["in", "inch", "inches"]:
            self.unit_multiplication_factor=2.54
        elif self.unit.lower() in ["ft", "foot", "feet"]:
            self.unit_multiplication_factor=30.479 
        
    def representation_scale(self):
        s=""
        # scale x
        if self.scale[0]!=1:
            s+="\nX-SCALE=({0},   0)".format(format(self.scale[0], "+22.15E"))
        # scale y
        if self.scale[1]!=1:
            s+="\nY-SCALE=({0},   0)".format(format(self.scale[1], "+22.15E"))
        # scale z
        if self.scale[2]!=1:
            s+="\nZ-SCALE=({0},   0)".format(format(self.scale[2], "+22.15E"))
        return s       
        
    def representation_rotation(self):
        s=""
        # omega
        if self.rotation[0]:
            s+="\n  OMEGA=({0},   0) {1}".format(format(self.rotation[0], "+22.15E"), self.angle)
        # theta
        if self.rotation[1]:
            s+="\n  THETA=({0},   0) {1}".format(format(self.rotation[1], "+22.15E"), self.angle)   
        # phi
        if self.rotation[2]:
            s+="\n    PHI=({0},   0) {1}".format(format(self.rotation[2], "+22.15E"), self.angle)
        return s

    def representation_translation(self):
        s=""
        # shift x
        if self.translation[0]:
            s+="\nX-SHIFT=({0},   0)".format(format(self.translation[0]*self.unit_multiplication_factor, "+22.15E"))
        # shift y
        if self.translation[1]:
            s+="\nY-SHIFT=({0},   0)".format(format(self.translation[1]*self.unit_multiplication_factor, "+22.15E"))
        # shift z
        if self.translation[2]:
            s+="\nZ-SHIFT=({0},   0)".format(format(self.translation[2]*self.unit_multiplication_factor, "+22.15E"))
        return s


class AdvancedElement(Element):
    def __init__(self, label, material, comment="", **kwargs):
        super().__init__(label, comment, **kwargs)
        self.material=int(material)
        self.surfaces=kwargs.get("surfaces", [])
        self.bodies=kwargs.get("bodies", [])
        self.modules=kwargs.get("modules", [])
        
    @abstractmethod
    def __str__(self):
        pass
        
    def representation_elements(self):
        s=""
        for surface in self.surfaces:
            if issubclass(type(surface[0]), Base):
                s+="\nSURFACE ({0}), SIDE POINTER=({1})".format(surface[0].label, str(surface[1]) if surface[1]<0 else " "+str(surface[1]))
            else:
                s+="\nSURFACE ({0}), SIDE POINTER=({1})".format(("    "+surface[0])[-4:], str(surface[1]) if surface[1]<0 else " "+str(surface[1]))
        # bodies
        for body in self.bodies:
            if issubclass(type(body), Base):
                s+="\nBODY    ({0})".format(body.label)
            else:
                s+="\nBODY    ({0})".format(("    "+body)[-4:])
        # modules
        for module in self.modules:
            if issubclass(type(module), Base):
                s+="\nMODULE  ({0})".format(module.label)
            else:
                s+="\nMODULE  ({0})".format(("    "+module)[-4:])
        return s       
        



__author__ = "Antonio Serrano"
__email__ = "toni.sm91@gmail.com"
__version__ = "0.0.2"
__license__ = "MIT License"

__all__ = ["GeometryDefinition"]

import os
try:
    from . import blocks
except:
    import blocks

# GEOMETRY-DEFINITION MANAGER

class GeometryDefinition():
    def __init__(self, description="", unit="cm", angle="DEG"):
        self.definition=[]
        self.description=description
        self.unit=unit
        self.angle=angle
        self.void_inner_volume_factor=1
        
        self._sm_characters=[120, 97, 97, 96]
 
    # surfaces
    def surface(self, label=None, indices=(1,1,1,1,1), starred=False, comment="", **kwargs):
        kwargs["unit"]=kwargs.get("unit", self.unit)
        kwargs["angle"]=kwargs.get("angle", self.angle)
        if label is None:
            label=self._sm_label()
        self.definition.append(blocks.Surface(label, indices, starred, comment, **kwargs))
        return self.definition[-1]
        
    def surface_implicit_form(self, label=None, indices=(0,0,0,0,0), starred=False, comment="", **kwargs):
        kwargs["unit"]=kwargs.get("unit", self.unit)
        kwargs["angle"]=kwargs.get("angle", self.angle)
        if label is None:
            label=self._sm_label()
        self.definition.append(blocks.Surface(label, indices, starred, comment, **kwargs))
        return self.definition[-1]

    def surface_plane(self, label=None, indices=(0,0,0,1,0), starred=False, comment="", **kwargs):
        kwargs["unit"]=kwargs.get("unit", self.unit)
        kwargs["angle"]=kwargs.get("angle", self.angle)
        if label is None:
            label=self._sm_label()
        self.definition.append(blocks.Surface(label, indices, starred, comment, **kwargs))
        return self.definition[-1]
            
    def surface_sphere(self, label=None, indices=(1,1,1,0,-1), starred=False, comment="", **kwargs):
        kwargs["unit"]=kwargs.get("unit", self.unit)
        kwargs["angle"]=kwargs.get("angle", self.angle)
        if label is None:
            label=self._sm_label()
        self.definition.append(blocks.Surface(label, indices, starred, comment, **kwargs))
        return self.definition[-1]

    def surface_cylinder(self, label=None, indices=(1,1,0,0,-1), starred=False, comment="", **kwargs):
        kwargs["unit"]=kwargs.get("unit", self.unit)
        kwargs["angle"]=kwargs.get("angle", self.angle)
        if label is None:
            label=self._sm_label()
        self.definition.append(blocks.Surface(label, indices, starred, comment, **kwargs))
        return self.definition[-1]
            
    def surface_hyperbolic_cylinder(self, label=None, indices=(1,-1,0,0,-1), starred=False, comment="", **kwargs):
        kwargs["unit"]=kwargs.get("unit", self.unit)
        kwargs["angle"]=kwargs.get("angle", self.angle)
        if label is None:
            label=self._sm_label()
        self.definition.append(blocks.Surface(label, indices, starred, comment, **kwargs))
        return self.definition[-1]
            
    def surface_cone(self, label=None, indices=(1,1,-1,0,0), starred=False, comment="", **kwargs):
        kwargs["unit"]=kwargs.get("unit", self.unit)
        kwargs["angle"]=kwargs.get("angle", self.angle)
        if label is None:
            label=self._sm_label()
        self.definition.append(blocks.Surface(label, indices, starred, comment, **kwargs))
        return self.definition[-1]

    def surface_one_sheet_hyperboloid(self, label=None, indices=(1,1,-1,0,-1), starred=False, comment="", **kwargs):
        kwargs["unit"]=kwargs.get("unit", self.unit)
        kwargs["angle"]=kwargs.get("angle", self.angle)
        if label is None:
            label=self._sm_label()
        self.definition.append(blocks.Surface(label, indices, starred, comment, **kwargs))
        return self.definition[-1]
            
    def surface_two_sheet_hyperboloid(self, label=None, indices=(1,1,-1,0,1), starred=False, comment="", **kwargs):
        kwargs["unit"]=kwargs.get("unit", self.unit)
        kwargs["angle"]=kwargs.get("angle", self.angle)
        if label is None:
            label=self._sm_label()
        self.definition.append(blocks.Surface(label, indices, starred, comment, **kwargs))
        return self.definition[-1]

    def surface_paraboloid(self, label=None, indices=(1,1,0,-1,0), starred=False, comment="", **kwargs):
        kwargs["unit"]=kwargs.get("unit", self.unit)
        kwargs["angle"]=kwargs.get("angle", self.angle)
        if label is None:
            label=self._sm_label()
        self.definition.append(blocks.Surface(label, indices, starred, comment, **kwargs))
        return self.definition[-1]
            
    def surface_parabolic_cylinder(self, label=None, indices=(1,0,0,-1,0), starred=False, comment="", **kwargs):
        kwargs["unit"]=kwargs.get("unit", self.unit)
        kwargs["angle"]=kwargs.get("angle", self.angle)
        if label is None:
            label=self._sm_label()
        self.definition.append(blocks.Surface(label, indices, starred, comment, **kwargs))
        return self.definition[-1]
            
    def surface_hyperbolic_paraboloid(self, label=None, indices=(1,-1,0,-1,0), starred=False, comment="", **kwargs):  
        kwargs["unit"]=kwargs.get("unit", self.unit)
        kwargs["angle"]=kwargs.get("angle", self.angle)
        if label is None:
            label=self._sm_label()
        self.definition.append(blocks.Surface(label, indices, starred, comment, **kwargs))
        return self.definition[-1]

    # body
    def body(self, label=None, material=0, comment="", **kwargs):
        kwargs["unit"]=kwargs.get("unit", self.unit)
        kwargs["angle"]=kwargs.get("angle", self.angle)
        if label is None:
            label=self._sm_label()
        self.definition.append(blocks.Body(label, material, comment, **kwargs))
        return self.definition[-1]

    # module
    def module(self, label=None, material=0, comment="", **kwargs):
        kwargs["unit"]=kwargs.get("unit", self.unit)
        kwargs["angle"]=kwargs.get("angle", self.angle)
        if label is None:
            label=self._sm_label()
        self.definition.append(blocks.Module(label, material, comment, **kwargs))
        return self.definition[-1]
        
    # clone
    def clone(self, label=None, module="", comment="", **kwargs):
        kwargs["unit"]=kwargs.get("unit", self.unit)
        kwargs["angle"]=kwargs.get("angle", self.angle)
        if label is None:
            label=self._sm_label()
        self.definition.append(blocks.Clone(label, module, comment, **kwargs))
        return self.definition[-1]
    
    # include
    def include(self, filename, starred=False, comment=""):
        self.definition.append(blocks.Include(filename, starred, comment))
        return self.definition[-1]
        
    # end
    def end(self):
        self.definition.append(blocks.End())
        return self.definition[-1]
        
    def export_definition(self, filename):
        with open(file=filename, mode='w') as file_object:
            s=""
            if self.description:
                s="\n"+self.description+"\n"
            for definition in self.definition:
                definition.set_void_inner_volume_factor(self.void_inner_volume_factor)
                s+="\n"+str(definition)
            try:
                if type(self.definition[-1])!=blocks.End:
                    s+="\n"+str(blocks.End())
                s+="\n"
            except IndexError:
                s="EMPTY GEOMETRY-DEFINITION"
            file_object.write(s)
        
    def show_void_inner_volumes(self, show=True):
        if show:
            self.void_inner_volume_factor=-1
        else:
            self.void_inner_volume_factor=1

    def _sm_label(self):
        # 52728 labels from 'XAAA' to 'ZZZZ'
        self._sm_characters[3]+=1
        if self._sm_characters[3]>122:
            self._sm_characters[3]=97
            self._sm_characters[2]+=1
            if self._sm_characters[2]>122:
                self._sm_characters[2]=97
                self._sm_characters[1]+=1
                if self._sm_characters[1]>122:
                    self._sm_characters[1]=97
                    self._sm_characters[0]+=1
                    if self._sm_characters[0]>122:
                        self._sm_characters[0]=120
        return "".join([chr(c).upper() for c in self._sm_characters])
        
    def __str__(self):
        s=""
        if self.description:
            s="\n"+self.description+"\n"
        for definition in self.definition:
            definition.set_void_inner_volume_factor(self.void_inner_volume_factor)
            s+="\n"+str(definition)
        try:
            if type(self.definition[-1])!=blocks.End:
                s+="\n"+str(blocks.End())
        except IndexError:
            return ""
        return s




if __name__=="__main__":
    
    g=GeometryDefinition("The pythonic champagne glass", unit="inch", angle="rad")
    
    s1=g.surface(starred=True)
    s2=g.surface(indices=(1,0,1,0,1), scale=(2,3,4), rotation=(5,6,7), translation=(8,9,1))
    s3=g.surface(indices=(1,0,1,0,1), xscale=20, yscale=30, zscale=40, omega=50, theta=60, phi=70, xshift=80, yshift=90, zshift=100, angle="deg")
    
    b1=g.body("B1", material=-100, comment="body number 1")
    b2=g.body("B2", material=-200, surfaces=[(s1, 1), (s2, -1)], bodies=[b1], comment="body number 2")
    
    m1=g.module(material=3, surfaces=[(s1, 1), (s2, -1), (s3, 1)], bodies=["B2"], modules=["M2"], scale=(2,3,4), rotation=(5,6,7), translation=(8,9,1), angle="deg", comment="module number 1")
    m2=g.module("M2", material=4, surfaces=[(s1, 1), (s2, -1), (s3, 1)], bodies=[b2], modules=[m1], xscale=20, yscale=30, zscale=40, omega=50, theta=60, phi=70, xshift=80, yshift=90, zshift=100, comment="module number 2")
    m3=g.module("M3", material=5, comment="module number 3")
    
    c1=g.clone("C1", m1, comment="clone number 1")
    c2=g.clone("C2", m2, scale=(2,3,4), rotation=(5,6,7), translation=(8,9,1), comment="clone number 2")
    c3=g.clone("C3", "M3", xscale=20, yscale=30, zscale=40, omega=50, theta=60, phi=70, xshift=80, yshift=90, zshift=100, unit="cm", angle="rad", comment="clone number 3")
    
    f1=g.include("filename1.test", comment="non starred file")
    f2=g.include("filename2.test", starred=True, comment="starred file")
    
    e=g.end()
    
    g.show_void_inner_volumes(False)
    
    print(g)
    g.export_definition("test")
    


from . import base

class Clone(base.Element):
    """
    Clonation of modules according to the next formatted text lines

    0000000000000000000000000000000000000000000000000000000000000000
    CLONE   (  A4) comment: copies one module and moves it
    MODULE  (  A4) original module
    1111111111111111111111111111111111111111111111111111111111111111
      OMEGA=(         E22.15       ,  I4) DEG (DEFAULT=0.0)
      THETA=(         E22.15       ,  I4) DEG (DEFAULT=0.0)
        PHI=(         E22.15       ,  I4) DEG (DEFAULT=0.0)
    X-SHIFT=(         E22.15       ,  I4) (DEFAULT=0.0)
    Y-SHIFT=(         E22.15       ,  I4) (DEFAULT=0.0)
    Z-SHIFT=(         E22.15       ,  I4) (DEFAULT=0.0)
    """

    def __init__(self, label, module, comment="", **kwargs):
        super().__init__(label, comment, **kwargs)
        if issubclass(type(module), base.Base):
            self.module=module.label
        else:
            self.module=("    "+module)[-4:]
        
    def __str__(self):
        s="0000000000000000000000000000000000000000000000000000000000000000\n"
        s+="CLONE   ({0}) {1}\n".format(self.label, self.comment)
        s+="MODULE  ({0})".format(self.module)
        # rotation and translation
        if self.rotation!=[0,0,0] or self.translation!=[0,0,0]:
            s+="\n1111111111111111111111111111111111111111111111111111111111111111"
        # rotation
        s+=self.representation_rotation()
        # translation
        s+=self.representation_translation()
        return s
